"""Test calculation of the minimal *mass* difference required to resolve two isotopic species
of a labeled chemical.

The minimal *mass* difference required to separate two isotopic species of a labeled chemical
correspond to the minimal difference of mz to distinguish two MS peaks at the operating resolution 
multiplied by the charge. The minimal mz difference is calculated at the operating 
resolution using 'resolution_formula' of orbitrap (Su et al., 2017) at the mz of the labeled chemical.
"""

import numpy as np
import pytest
import isocor as hrcor
import math

@pytest.mark.parametrize("data", [{"formula": "C20",
                                   "nC": 20,
                                   "resolution": 10000,
                                   "charge": 1},
                                   {"formula": "C20",
                                   "nC": 20,
                                   "resolution": 100000,
                                   "charge": 1},
                                   {"formula": "C20",
                                   "nC": 20,
                                   "resolution": 100000,
                                   "charge": 2}])
def test_m_min_orbitrap(data, data_iso):
    """Calculation of the minimal *mass* difference (m_min) required to resolve two isotopic species
    of a labeled chemical at different resolution (10000, 100000) & charge (1, 2) on an *Orbitrap*.
    m_min calculated by IsoCor is compared to the expected value (provided in the fixture) and to the
    theoretical value (equation provided in the code).
    This is checked for correctors instantiated through the Factory or directly as HighResMetaboliteCorrector. 
    """
    charge = data["charge"]
    mw = data["nC"]*float(data_iso["C"]["mass"][0])
    mz = mw/charge
    # Theoretical mmin 
    m_min_th = 1.66 * (mw/charge)**(3/2)/(data["resolution"]*math.sqrt(400))*charge
    # Estimated mmin
    #    when metabolite is instantiated directly as HighResMetaboliteCorrector
    metabolite_hr = hrcor.HighResMetaboliteCorrector(data["formula"], "13C", data_isotopes=data_iso,
                                                  resolution=data["resolution"],
                                                  resolution_formula_code="orbitrap",
                                                  derivative_formula=None,
                                                  mz_of_resolution=400,
                                                  correct_NA_tracer=False,
                                                  tracer_purity=[0.0, 1.0],
                                                  charge=charge)
    m_min_isocor_hr = float(metabolite_hr._correction_limit)
    #    when metabolite is instantiated through the Factory
    metabolite_factory = hrcor.MetaboliteCorrectorFactory(data["formula"], "13C", data_isotopes=data_iso,
                                                  resolution=data["resolution"],
                                                  resolution_formula_code="orbitrap",
                                                  derivative_formula=None,
                                                  mz_of_resolution=400,
                                                  correct_NA_tracer=False,
                                                  tracer_purity=[0.0, 1.0],
                                                  charge=charge)
    m_min_isocor_factory = float(metabolite_factory._correction_limit)
    # Compare estimated, theoretical and expected minimal mass difference to resolve two isotopic species
    np.testing.assert_allclose(m_min_isocor_hr, m_min_th, rtol=1e-7, atol=1e-7)
    np.testing.assert_allclose(m_min_isocor_factory, m_min_th, rtol=1e-7, atol=1e-7)



"""Test the entire high-resolution correction process at high-resolution.

Comparison of
mass fractions corrected by IsoCor to theoretical isotopologues distributions
(with algebraic equations provided in the code for manual check).
"""

import numpy as np
import pytest
import isocor as hrcor

# Tests the entire correction process at high resolution
# (here pXY represents natural abundance of isotope X of element Y)


@pytest.mark.parametrize("data", [{"formula": "C2H2",
                                   "resolution_at_400": 1e3,
                                   "tracer": "13C",
                                   "v_expected": "[1., 0., 0.]",
                                   "v_measured": "[p1H * p1H, p1H * p2H * 2, p2H * p2H]",
                                   "correct_NA_tracer": False,
                                   "tracer_purity": [0.0, 1.0]},
                                  {"formula": "C2H2",
                                   "resolution_at_400": 1e4,
                                   "tracer": "13C",
                                   "v_expected": "[0.2, 0.3, 0.5]",
                                   "v_measured": "[0.2, 0.3, 0.5]",
                                   "correct_NA_tracer": False,
                                   "tracer_purity": [0.0, 1.0]},
                                  {"formula": "C2H2P10",
                                   "resolution_at_400": 1e4,
                                   "tracer": "13C",
                                   "v_expected": "[1., 0., 0.]",
                                   "v_measured": "[p1H * p1H, p1H * p2H * 2, p2H * p2H]",
                                   "correct_NA_tracer": False,
                                   "tracer_purity": [0.0, 1.0]},
                                  {"formula": "C2H2P10",
                                   "resolution_at_400": 1e6,
                                   "tracer": "13C",
                                   "v_expected": "[0.2, 0.3, 0.5]",
                                   "v_measured": "[0.2, 0.3, 0.5]",
                                   "correct_NA_tracer": False,
                                   "tracer_purity": [0.0, 1.0]},
                                  {"formula": "C2H2P10",
                                   "resolution_at_400": 1e4,
                                   "tracer": "13C",
                                   "v_expected": "[0.7, 0.2, 0.1]",
                                   "v_measured": "[0.7*p1H * p1H, 0.7*p1H * p2H * 2+0.2*p1H * p1H, 0.7*p2H * p2H+0.2*p1H * p2H * 2+0.1*p1H * p1H]",
                                   "correct_NA_tracer": False,
                                   "tracer_purity": [0.0, 1.0]},
                                  {"formula": "C2H2P10",
                                   "resolution_at_400": 1e6,
                                   "tracer": "13C",
                                   "v_expected": "[0.7, 0.2, 0.1]",
                                   "v_measured": "[0.7, 0.2, 0.1]",
                                   "correct_NA_tracer": False,
                                   "tracer_purity": [0.0, 1.0]},
                                  {"formula": "OH2",
                                   "resolution_at_400": 1e2,
                                   "tracer": "18O",
                                   "v_expected": "[0.6, 0.4]",
                                   "v_measured": "[0.6*p1H**2, 0.6*p2H**2+0.4*p1H**2]",
                                   "correct_NA_tracer": False,
                                   "tracer_purity": [0.0, 0.0, 1.0]},
                                  {"formula": "OH2",
                                   "resolution_at_400": 1e2,
                                   "tracer": "17O",
                                   "v_expected": "[0.6, 0.4]",
                                   "v_measured": "[0.6*p1H**2, 0.6*p1H*p2H*2+0.4*p1H**2]",
                                   "correct_NA_tracer": False,
                                   "tracer_purity": [0.0, 1.0, 0.0]},
                                  {"formula": "OH2",
                                   "resolution_at_400": 1e6,
                                   "tracer": "18O",
                                   "v_expected": "[0.6, 0.4]",
                                   "v_measured": "[0.6, 0.4]",
                                   "correct_NA_tracer": False,
                                   "tracer_purity": [0.0, 0.0, 1.0]},
                                  {"formula": "OH2",
                                   "resolution_at_400": 1e6,
                                   "tracer": "17O",
                                   "v_expected": "[0.6, 0.4]",
                                   "v_measured": "[0.6, 0.4]",
                                   "correct_NA_tracer": False,
                                   "tracer_purity": [0.0, 1.0, 0.0]},
                                  {"formula": "C2H2P10",
                                   "resolution_at_400": 1e4,
                                   "tracer": "13C",
                                   "v_expected": "[0.7, 0.2, 0.1]",
                                   "v_measured": "[0.7*p1H*p1H*p12C**2, 0.7*p1H*p2H*2*p12C**2 + 0.7*p1H*p1H*p12C*p13C*2 + 0.2*p1H*p1H*p12C, 0.7*p2H*p2H*p12C**2 + 0.7*p1H*p2H*2*p12C*p13C*2 + 0.7*p1H*p1H*p13C**2 +0.2*p1H*p1H*p13C + 0.2*p1H*p2H*2*p12C + 0.1*p1H*p1H]",
                                   "correct_NA_tracer": True,
                                   "tracer_purity": [0.0, 1.0]},
                                  {"formula": "C2H2P10",
                                   "resolution_at_400": 1e4,
                                   "tracer": "13C",
                                   "v_expected": "[0.7, 0.2, 0.1]",
                                   "v_measured": "[0.7*p1H**2 + 0.2*p1H**2*pur12C + 0.1*p1H**2*pur12C**2, 0.7*p1H*p2H*2 + 0.2*p1H**2*pur13C + 0.2*p1H*p2H*2*pur12C + 0.1*p1H**2*pur12C*pur13C*2 + 0.1*p1H*p2H*2*pur12C**2, 0.7*p2H**2 + 0.2*p1H*p2H*2*pur13C + 0.2*p2H**2*pur12C + 0.1*p1H**2*pur13C**2 + 0.1*p1H*p2H*2*pur12C*pur13C*2 + 0.1*p2H**2*pur12C**2]",
                                   "correct_NA_tracer": False,
                                   "tracer_purity": [0.3, 0.7]}])
def test_high_res_correction(data, data_iso):
    """Correction of non-tracer elements depending on resolution and formula.

    Measured values should be corrected for the incorporation of natural isotopes of
    non-tracers elements if the resolution was not high enough to distinguish them.
    The resolution depends on m/z ratio and thus the formula of the metabolite.
    The higher the mass, the higher the resolution should be used to distinguish all
    isotopomers and avoid to need correction.
    """
    formula = data["formula"]
    resolution = data["resolution_at_400"]
    # Observed abundance of each isotopomers given the natural abundance
    # of the H isotopes
    p1H = data_iso["H"]["abundance"][0]
    p2H = data_iso["H"]["abundance"][1]
    p12C = data_iso["C"]["abundance"][0]
    p13C = data_iso["C"]["abundance"][1]
    pur12C = data["tracer_purity"][0]
    pur13C = data["tracer_purity"][1]
    v_measured = eval(data["v_measured"])
    # Expected correction
    v_expected = eval(data["v_expected"])
    # Perform correction
    metabolite = hrcor.HighResMetaboliteCorrector(formula, data["tracer"], data_isotopes=data_iso,
                                                  resolution=resolution,
                                                  resolution_formula_code="orbitrap",
                                                  derivative_formula=None,
                                                  mz_of_resolution=400,
                                                  correct_NA_tracer=data["correct_NA_tracer"],
                                                  tracer_purity=data["tracer_purity"],
                                                  charge=1)
    _, v_corrected, _, _ = metabolite.correct(v_measured)
    # Compare corrected vs. expected data
    np.testing.assert_allclose(v_corrected, v_expected, rtol=1e-7)



"""Test the calculation of high-resolution isotopic clusters (for metabolites with
different number of elements and isotopes) against a brute force implementation.
"""

import math
import itertools as it
import pytest
import numpy as np
import isocor as hrcor


def get_isoclust_bruteforce(formula, data_iso):
    """Return the isotopic cluster for a compound using a brute-force method.

    This method is based on systematic computation of all isotopic species of the molecule.
    It should be easier to understand than the optimized version but is much slower.
    """
    def npermutations(v):
        dem = 1
        for x in set(v):
            dem *= math.factorial(v.count(x))
        return math.factorial(len(v)) / dem
    isotopic_cluster = {}  # {isotopologue_mass: isotopologue_proba, ...}
    # For each element, gather the possible combinations of its isotopes
    element_blocks = []   # list of iterators over isotopes combinations for one element
    elements_order = []  # e.g.: ["C", "N", "H"] states thats "C" block is at index 0
    for element in formula:
        # number of natural isotopes
        n_isotopes = len(data_iso[element]["mass"])
        iter_iso_combi = it.combinations_with_replacement(range(n_isotopes),  # allowed isotopes (index of data_iso)
                                                          formula[element])   # number of this element in formula
        elements_order.append(element)
        element_blocks.append(iter_iso_combi)
    # Make an iterator over all isotopologues
    # NB: simply take one combination/block for each element
    iter_isotopologues = it.product(*element_blocks)
    # Process each isotopologue
    for isotopologue in iter_isotopologues:
        # Compute the mass of the isotopologue
        mass_by_block = []
        for idx_element, combi_isotopes in enumerate(isotopologue):
            element = elements_order[idx_element]
            isotope_mass = [data_iso[element]["mass"][idx_iso]
                            for idx_iso in combi_isotopes]
            mass_by_block.append(sum(isotope_mass))
        mass_isotopologue = sum(mass_by_block)
        # Compute the expected abundance for this isotopologue
        n_isotopomers_by_block = []  # number of isotopomers caused by each element_block
        listof_block_proba = []      # proba of each element_block
        for idx_element, combi_isotopes in enumerate(isotopologue):
            n_isotopomers_by_block.append(npermutations(combi_isotopes))
            element = elements_order[idx_element]  # e.g. "C"
            block_proba = np.prod(
                [data_iso[element]["abundance"][idx_iso] for idx_iso in combi_isotopes])
            listof_block_proba.append(block_proba)
        proba_isotopologue = np.prod(
            n_isotopomers_by_block) * np.prod(listof_block_proba)
        # Save
        assert mass_isotopologue not in isotopic_cluster
        isotopic_cluster[mass_isotopologue] = proba_isotopologue
    return isotopic_cluster


@pytest.mark.parametrize("formula", [{"C": 1, "N": 1},
                                     {"C": 1, "O": 1},
                                     {"C": 1, "N": 2},
                                     {"C": 1, "O": 2},
                                     {"C": 1, "N": 20, "H": 20, "O": 20, "P": 4},
                                     {"C": 20, "N": 20, "H": 20, "O": 20, "P": 4}])
def test_isoclust_against_bruteforce(formula, data_iso, usr_tolerance):
    """Check the high-resolution isotopic clusters generation against a brute force implementation."""
    str_formula = "".join(["{}{}".format(k, v) for k, v in formula.items()])
    # Get the clusters (without tracers)
    try:
        del formula["C"]
    except:
        pass
    ic_bruteforce = get_isoclust_bruteforce(formula, data_iso)
    metabolite = hrcor.HighResMetaboliteCorrector(str_formula, '13C', 1e42, 400,
                                                  resolution_formula_code="orbitrap",
                                                  derivative_formula=None,
                                                  tracer_purity=None,
                                                  data_isotopes=data_iso,
                                                  correct_NA_tracer=False,
                                                  charge=1)
    metabolite.threshold_p = None
    ic_optimized = metabolite.get_isotopic_cluster()
    # Check that all isotopomeres are taken into account
    np.testing.assert_allclose(
        sum(ic_bruteforce.values()), 1., rtol=usr_tolerance)
    np.testing.assert_allclose(
        sum(ic_optimized.values()), 1., rtol=usr_tolerance)
    # Check that both methods have the same number of isotopologues
    assert len(ic_bruteforce) == len(ic_optimized)
    # Finally, check each mass (key) and each abundance (value)
    mass_optimized = [float(x) for x in sorted(ic_optimized.keys())]
    mass_bruteforce = [float(x) for x in sorted(ic_bruteforce.keys())]
    np.testing.assert_allclose(
        mass_optimized, mass_bruteforce, rtol=usr_tolerance)
    abun_optimized = [float(x) for x in sorted(ic_optimized.values())]
    abun_bruteforce = [float(x) for x in sorted(ic_bruteforce.values())]
    np.testing.assert_allclose(
        abun_optimized, abun_bruteforce, rtol=usr_tolerance)



"""Test high-resolution correction matrix when some or all isotopic species are resolved from the tracer isotopologues.

In this situation, the contribution of only some isotopic species should be removed,
depending on the resolution.
The high-resolution correction matrix calculated by IsoCor is compared to a theoretical correction matrix
(with algebraic equations provided in the code for manual check).
"""

import numpy as np
import pytest
import isocor as hrcor

# Test the correction matrix by comparison with theoretical matrices expressed using explicit functions (here pXY represents natural
# abundance of isotope X of element Y), which can be easily verified. The examples provided are simple enough for manual verification,
# yet they represent the many situations that can be encountered in the real world. Each situation is detailed in the field "comment".

# Test construction of the correction matrix at high resolution
# tests below are designed to evaluate the correction of high-resolution MS data where some isotopic species are resolved from the tracer-isotopologues


@pytest.mark.parametrize("data", [{"comment": "13C tracer, no derivative, no NA of the tracer element, no purity of the tracer element, tracer with 2 isotopes, correct for NA of 1 elements with 3 isotopes, 17O and 18O are resolved",
                                   "formula": "C2O2P10",
                                   "tracer": "13C",
                                   "derivative_formula": "",
                                   "resolution": 1e6,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[p16O**2, 0.0, 0.0], [0.0, p16O**2, 0.0], [0.0, 0.0, p16O**2]]",
                                   "tracer_purity": [0.0, 1.0]},
                                  {"comment": "13C tracer, no derivative, no NA of the tracer element, no purity of the tracer element, tracer with 2 isotopes, correct for NA of 1 elements with 2 isotopes",
                                   "formula": "C2H2P10",
                                   "tracer": "13C",
                                   "derivative_formula": "",
                                   "resolution": 1e6,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[p1H**2, 0.0, 0.0], [0.0, p1H**2, 0.0], [0.0, 0.0, p1H**2]]",
                                   "tracer_purity": [0.0, 1.0]},
                                  {"comment": "13C tracer, no derivative, with NA of the tracer element, no purity of the tracer element, tracer with 2 isotopes, correct for NA of 1 elements with 2 isotopes",
                                   "formula": "C2H2P10",
                                   "tracer": "13C",
                                   "derivative_formula": "",
                                   "resolution": 1e6,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": True,
                                   "expected_matrix": "[[p1H**2*p12C**2, 0.0, 0.0], [2*p12C*p13C*p1H**2, p1H**2*p12C, 0.0], [p1H**2*p13C**2, p1H**2*p13C, p1H**2]]",
                                   "tracer_purity": [0.0, 1.0]},
                                  {"comment": "13C tracer, no derivative, no NA of the tracer element, no purity of the tracer element, tracer with 2 isotopes, correct for NA of 2 elements with 2 isotopes",
                                   "formula": "C2H2N2P10",
                                   "tracer": "13C",
                                   "derivative_formula": "",
                                   "resolution": 1e6,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[p1H**2*p14N**2, 0.0, 0.0], [0.0, p1H**2*p14N**2, 0.0], [0.0, 0.0, p1H**2*p14N**2]]",
                                   "tracer_purity": [0.0, 1.0]},
                                  {"comment": "13C tracer, no derivative, no NA of the tracer element, with purity of the tracer element, tracer with 2 isotopes, correct for NA of 1 elements with 2 isotopes",
                                   "formula": "C2H2P10",
                                   "tracer": "13C",
                                   "derivative_formula": "",
                                   "resolution": 1e6,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[p1H**2, p1H**2*pur_12C, p1H**2*pur_12C*pur_12C], [0.0, p1H**2*pur_13C, p1H**2*pur_12C*pur_13C*2], [0.0, 0.0, p1H**2*pur_13C*pur_13C]]",
                                   "tracer_purity": [0.1, 0.9]},
                                  # tests for tracers having more than 2 isotopes
                                  {"comment": "17O tracer, no derivative, no NA of the tracer element, with purity of the tracer element, tracer with 2 isotopes, correct for NA of 1 elements with 2 isotopes",
                                   "formula": "O",
                                   "tracer": "17O",
                                   "derivative_formula": "",
                                   "resolution": 1e6,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": True,
                                   "expected_matrix": "[[p16O, 0.0], [p17O, 0.0]]",
                                   "tracer_purity": [0.0, 0.0, 1.0]},
                                  {"comment": "17O tracer, no derivative, NA of the tracer element, no purity of the tracer element, tracer with 3 isotopes, correct for NA of 0 elements",
                                   "formula": "O",
                                   "tracer": "17O",
                                   "derivative_formula": "",
                                   "resolution": 1e6,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": True,
                                   "expected_matrix": "[[p16O, 0.0], [p17O, 1.0]]",
                                   "tracer_purity": [0.0, 1.0, 0.0]},
                                  {"comment": "17O tracer, no derivative, no NA of the tracer element, no purity of the tracer element, tracer with 3 isotopes, correct for NA of 0 elements",
                                   "formula": "O1",
                                   "tracer": "17O",
                                   "derivative_formula": "",
                                   "resolution": 1e6,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[1.0, 0.0], [0.0, 1.0]]",
                                   "tracer_purity": [0.0, 1.0, 0.0]},
                                  {"comment": "17O tracer, no derivative, no NA of the tracer element, with purity of the tracer element, tracer with 3 isotopes, correct for NA of 0 elements",
                                   "formula": "O1",
                                   "tracer": "17O",
                                   "derivative_formula": "",
                                   "resolution": 1e6,
                                   "mz_of_resolution": 400,
                                   "correct_NA_tracer": False,
                                   "expected_matrix": "[[1.0, 0.1], [0.0, 0.9]]",
                                   "tracer_purity": [0.1, 0.9, 0.0]}])
def test_high_res_cor_matrix(data, data_iso, usr_tolerance):
    """Correction of high-resolution MS data.

    Test calculates the correction matrix of labeled metabolites containing C, H, N and O elements,
    and compares this matrix to a theoretical matrix.
    """
    # perform correction
    metabolite = hrcor.HighResMetaboliteCorrector(data["formula"], data["tracer"],
                                                  data_isotopes=data_iso,
                                                  correct_NA_tracer=data["correct_NA_tracer"],
                                                  resolution_formula_code = "orbitrap",
                                                  tracer_purity=data["tracer_purity"],
                                                  resolution=data["resolution"],
                                                  mz_of_resolution=data["mz_of_resolution"],
                                                  derivative_formula=data["derivative_formula"],
                                                  charge=1)
    correction_matrix = metabolite.compute_correction_matrix()
    # Test against expected matrix
    p1H = data_iso["H"]["abundance"][0]
    p2H = data_iso["H"]["abundance"][1]
    p12C = data_iso["C"]["abundance"][0]
    p13C = data_iso["C"]["abundance"][1]
    p14N = data_iso["N"]["abundance"][0]
    p15N = data_iso["N"]["abundance"][1]
    p16O = data_iso["O"]["abundance"][0]
    p17O = data_iso["O"]["abundance"][1]
    p18O = data_iso["O"]["abundance"][2]
    pur_12C = data["tracer_purity"][0]
    pur_13C = data["tracer_purity"][1]
    pur1H = data["tracer_purity"][0]
    pur2H = data["tracer_purity"][1]
    expected_matrix = eval(data["expected_matrix"])
    np.testing.assert_allclose(
        correction_matrix, expected_matrix, rtol=usr_tolerance)



import isocor.ui.isocorgui
import isocor.ui.isocorcli
import sys

# start CLI if arguments provided, otherwise switch to GUI
if len(sys.argv) > 1:
    isocor.ui.isocorcli.start_cli()
else:
    isocor.ui.isocorgui.start_gui()
